import ck_bandit.base as ck_base
import ck_bandit.utils as ck_utils
import sklearn.gaussian_process.kernels as sk_kern
import numpy as np
import math
import scipy.stats
import itertools
from typing import Tuple, List
import scipy.special
from multiprocessing import Pool


def cvar_lognormal(mu: float, sigma: float, alpha: float):
    q = scipy.stats.norm.ppf(alpha)
    return math.exp(mu + sigma**2 / 2.0) * (scipy.stats.norm.cdf(q - sigma)) / alpha


def cvar_normal(mu, sigma, alpha):
    q = scipy.stats.norm.ppf(alpha)
    phi = math.exp(-q**2/2.0)/math.sqrt(2*math.pi)
    return mu - sigma * phi / alpha


def mv_normal(mu: float, sigma: float, c: float) -> float:
    return mu - sigma * sigma * c


def mv_lognormal(mu: float, sigma: float, c: float) -> float:
    s2 = sigma * sigma
    mean = math.exp(mu + s2 * 0.5)
    var = (math.exp(s2) - 1) * math.exp(2 * mu + s2)
    return mean - c * var


def test_normal_mu_sigma(x: float) -> Tuple[float, float]:
    return math.cos(2*math.pi * x), (1.5 + math.sin(2*math.pi * x))


def mucfs_mut_pts_sigma_cfs_siga_pts(num_pts_rand, kern, dim, rng):
    mu_cfs, mu_pts = ck_utils.random_func_rkhs_of_unit_norm(
        num_pts_rand, kern, lbs=np.zeros(dim), ubs=np.ones(dim), rng=rng)
    sigma_cfs, sigma_pts = ck_utils.random_func_rkhs_of_unit_norm(
        num_pts_rand, kern, lbs=np.zeros(dim), ubs=np.ones(dim), rng=rng)
    return mu_cfs, mu_pts, sigma_cfs, sigma_pts


class RandNormalLikeBaseEnv(ck_base.Environment):
    def __init__(self, kern: sk_kern.Kernel, rng: np.random.Generator, seed: int,
                 dim: int, kern_name: str, npts_each_dim=10, num_pts_rand=100, sigma_min=1e-3):
        self.rng = rng
        self.seed = seed
        self.dim = dim
        self.kern = kern
        self.kern_name = kern_name
        self.npts_each_dim = npts_each_dim
        itrs = [np.linspace(0.0, 1.0, num=self.npts_each_dim + 1) for _ in range(dim)]
        self._arms = np.array(list(itertools.product(*itrs)))

        args = (num_pts_rand, kern, dim, rng)
        mu_cfs, mu_pts, sigma_cfs, sigma_pts = mucfs_mut_pts_sigma_cfs_siga_pts(*args)

        self.mu_cfs = mu_cfs
        self.mu_pts = mu_pts
        self.sigma_cfs = sigma_cfs
        self.sigma_pts = sigma_pts
        self.sigma_min = sigma_min

        mus, sigmas = self.mus_and_sigmas(self._arms,)

        self._mus = mus
        self._sigmas = sigmas

    def arms(self) -> np.ndarray:
        return self._arms

    def mu_and_sigma(self, arm: np.ndarray) -> Tuple[float, float]:
        arm = np.array(arm)
        mu = self.kern(arm, self.mu_pts) @ self.mu_cfs
        mu = mu.reshape((1,))[0]
        sigma = self.kern(arm, self.sigma_pts) @ self.sigma_cfs
        sigma = sigma.reshape((1,))[0]
        sigma = math.sqrt(sigma * sigma + self.sigma_min)
        return mu, sigma

    def mus_and_sigmas(self, arms: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        n = arms.shape[0]
        mu_cfs = self.mu_cfs.reshape((self.mu_cfs.size, 1))
        mus = self.kern(arms, self.mu_pts) @ mu_cfs
        mus = mus.reshape((n,))
        sigma_cfs = self.sigma_cfs.reshape((self.sigma_cfs.size, 1))
        sigma = self.kern(arms, self.sigma_pts) @ sigma_cfs
        sigma = sigma.reshape((n,))
        sigma = np.sqrt(sigma * sigma + self.sigma_min)
        return mus, sigma

    def mean_vars(self, arms: np.ndarray, c: float) -> List[float]:
        return [self.mean_var(a, c) for a in arms]

    def best_mv_arm_and_reward(self, c: float) -> Tuple[np.ndarray, float]:
        arms = self.arms()
        mvs = self.mean_vars(arms, c)
        i = np.argmax(mvs)
        return (arms[i], mvs[i])


class RandNormalEnv(RandNormalLikeBaseEnv):
    def observation(self, arm: np.ndarray, rng: np.random.Generator) -> float:
        mu, sigma = self.mu_and_sigma(arm)
        obs = rng.normal(mu, sigma)
        return obs

    def cvar(self, arm: np.ndarray, alpha: float) -> float:
        mu, sigma = self.mu_and_sigma(arm)
        return cvar_normal(mu, sigma, alpha)

    def cvars(self, arms: np.ndarray, alpha: float) -> List[float]:
        mus, sigmas = self.mus_and_sigmas(arms)
        return [cvar_normal(mu, sigma, alpha) for mu, sigma in zip(mus, sigmas)]

    def best_arm_and_reward(self, alpha: float) -> Tuple[np.ndarray, float]:
        arms = self.arms()
        cvars = self.cvars(arms, alpha)
        i = np.argmax(cvars)
        return (arms[i], cvars[i])

    def name(self) -> str:
        return "rn{}d{}k{}".format(self.dim, self.seed, self.kern_name)

    def min_max_of_q(self, p: float) -> Tuple[float, float]:
        eri = scipy.special.erfinv(2 * (1 - p) - 1)
        mx = np.max(self._mus + math.sqrt(2) * self._sigmas * eri)

        eri = scipy.special.erfinv(2 * p - 1)
        mn = np.min(self._mus + math.sqrt(2) * self._sigmas * eri)
        return (mn, mx)

    def mean_var(self, arm: np.ndarray, c: float) -> float:
        m, s = self.mu_and_sigma(arm)
        return mv_normal(m, s, c)


class RandLogNormalEnv(RandNormalLikeBaseEnv):
    def observation(self, arm: np.ndarray, rng: np.random.Generator) -> float:
        mu, sigma = self.mu_and_sigma(arm)
        return rng.lognormal(mean=mu, sigma=sigma)

    def cvar(self, arm: np.ndarray, alpha: float) -> float:
        mu, sigma = self.mu_and_sigma(arm)
        return cvar_lognormal(mu, sigma, alpha)

    def cvars(self, arms: np.ndarray, alpha: float) -> List[float]:
        mus, sigmas = self.mus_and_sigmas(arms)
        return [cvar_lognormal(mu, sigma, alpha) for mu, sigma in zip(mus, sigmas)]

    def name(self) -> str:
        return "rln{}d{}k{}".format(self.dim, self.seed, self.kern_name)

    def best_arm_and_reward(self, alpha: float) -> Tuple[np.ndarray, float]:
        arms = self.arms()
        cvars = []
        for a in arms:
            mu, sigma = self.mu_and_sigma(a)
            cvars.append(cvar_lognormal(mu, sigma, alpha))
        i = np.argmax(cvars)
        return (arms[i], cvars[i])

    def min_max_of_q(self, p: float) -> Tuple[float, float]:
        eri = scipy.special.erfinv(2 * (1 - p) - 1)
        mx = np.max(self._mus + math.sqrt(2) * self._sigmas * eri)

        eri = scipy.special.erfinv(2 * p - 1)
        mn = np.min(self._mus + math.sqrt(2) * self._sigmas * eri)
        return (math.exp(mn), math.exp(mx))

    def mean_var(self, arm: np.ndarray, c: float) -> float:
        m, s = self.mu_and_sigma(arm)
        return mv_lognormal(m, s, c)


class ParamLogNormal1DEnv(ck_base.Environment):
    def __init__(self, narms=100):
        self.narms = narms

    def arms(self) -> np.ndarray:
        arms = np.linspace(0.0, 1.0, num=self.narms)
        return arms.reshape((self.narms, 1))

    def obvervation(self, arm: np.ndarray, rng: np.random.Generator) -> float:
        mu, sigma = test_normal_mu_sigma(arm[0])
        return rng.lognormal(mean=mu, sigma=sigma)

    def best_arm_and_reward(self, alpha: float) -> Tuple[np.ndarray, float]:
        arms = self.arms()
        cvars = []
        for a in arms:
            mu, sigma = test_normal_mu_sigma(a[0])
            cvars.append(cvar_lognormal(mu, sigma, alpha))
        i = np.argmax(cvars)
        return (arms[i], cvars[i])

    def name(self) -> str:
        return "lpn1d{}".format(self.narms)


class ParamNormal1DEnv(ck_base.Environment):
    def __init__(self, narms=100):
        self.narms = narms

    def arms(self) -> np.ndarray:
        arms = np.linspace(0.0, 1.0, num=self.narms)
        return arms.reshape((self.narms, 1))

    def observation(self, arm: np.ndarray, rng: np.random.Generator) -> float:
        mu, sigma = test_normal_mu_sigma(arm[0])
        return rng.normal(mean=mu, sigma=sigma)

    def best_arm_and_reward(self, alpha: float) -> Tuple[np.ndarray, float]:
        arms = self.arms()
        cvars = []
        for a in arms:
            mu, sigma = test_normal_mu_sigma(a[0])
            cvars.append(cvar_normal(mu, sigma, alpha))
        i = np.argmax(cvars)
        return (arms[i], cvars[i])

    def name(self) -> str:
        return "pn1d{}".format(self.narms)
