import pickle
import ck_bandit.run_policy as ck_run_pol
import ck_bandit.pols.gp_bandits as gp_bandits
import ck_bandit.pols.pols_misc as ck_pols_misc
from multiprocessing import Pool
import numpy as np
from typing import Dict
from ck_bandit.constants import *


def load_best_params():
    with open(BEST_PARAMS_PATH, "rb") as fp:
        return pickle.load(fp)


def cumreg_cvars_using_env_wrapper(dim: int, env_type: str, pol, nrounds, env_seed, alpha, is_eval=True) -> np.ndarray:
    if is_eval:
        obs_seed = EVAL_OBS_SEED
    else:
        obs_seed = PARAM_SEARCH_OBS_SEED
    if is_eval:
        env = ck_run_pol.ENV_FNS_EVAL[env_type](dim=dim, seed=env_seed)
    else:
        env = ck_run_pol.ENV_FNS_PARAM_SEARCH[env_type](dim=dim, seed=env_seed)
    return ck_run_pol.cumreg_cvars_using_env(pol, env, nrounds, obs_seed, alpha, norm_fac=1.0)


def cum_regret(env_type: str, alpha: float) -> Dict[str, np.ndarray]:
    '''
    Parameters
    ----------
    env_type: str should be 'normal' or 'log_normal' 
    alpha: float should be in [0.05, 0.1, 0.5, 0.9, 0.95]
    ----------
    
    Returns
    ----------
    Dict[str, np.ndarray]

    returns a dictionay whose set of keys is {'IGP-UCB', 'CVPKE-UCB'}
    and the corresponding value is a numpy array of shape (10, 1000).
    The (i, j) entry of the corresponding array represents 
    normalized cummulative CVaR regret up to round j 
    when the policy interatcts with the i-th synthetic environment.
    '''
    assert alpha in ALPHAS
    dim = DIMS[0]
    d = load_best_params()
    nrounds = NROUNDS_EVAL
    is_eval = True
    env_seeds = EVAL_SEEDS

    prmigp = d["igp"][(dim, env_type, alpha)]
    prmcvar = d["cvarucb"][(dim, env_type, alpha)]
    pols = [gp_bandits.IGPUCB(prmigp), gp_bandits.CVKMEUCB(prmcvar, alpha)]
    pol_names = ["IGP-UCB", "CVPKE-UCB"]

    res = {}

    norm_facs = []
    for env_seed in env_seeds:
        # compute cum_reg_for random
        with Pool(1) as pl:
            rand_pol = ck_pols_misc.RandomPolicy()
            args = (dim, env_type, rand_pol, nrounds, env_seed, alpha, is_eval)
            cum_reg_rand = pl.apply(cumreg_cvars_using_env_wrapper, args)
            norm_factor = cum_reg_rand[nrounds - 1]
            cum_reg_rand /= norm_factor
        norm_facs.append(norm_factor)

    for pol, pol_name in zip(pols, pol_names):
        cum_regs = []
        for env_seed, norm_fac in zip(env_seeds, norm_facs):
            with Pool(1) as pl:
                args = (dim, env_type, pol, nrounds, env_seed, alpha, is_eval)
                cum_reg = pl.apply(cumreg_cvars_using_env_wrapper, args)
                cum_reg /= norm_fac
            cum_regs.append(cum_reg)
        cum_regs = np.array(cum_regs)
        res[pol_name] = cum_regs

    return res
