from typing import Optional
import ck_bandit.base as ck_base
import ck_bandit.utils as ck_utils
import numpy as np
import sklearn.gaussian_process.kernels as sk_kern
import math


def sigmas(xs: np.ndarray, xhist: np.ndarray, kxs_kmat_inv: np.ndarray, kern: sk_kern.Kernel) -> np.ndarray:
    kxts = kern(xs, xhist)
    a = ck_utils.dots_of_arrays(kxs_kmat_inv, kxts)
    sigma2 = np.array([kern([x], [x])[0, 0] for x in xs]) - a
    return np.sqrt(sigma2)


def sigma_each(x: np.ndarray, xhist: np.ndarray, kmat_inv: np.ndarray, kern: sk_kern.Kernel) -> np.ndarray:
    kxs = kern([x], xhist)
    kxs = kxs.reshape((kxs.size,))
    res2 = kern([x], [x])[0, 0] - (kmat_inv @ kxs).dot(kxs)
    return np.sqrt(res2)


def mu_means(yhist: np.ndarray, kxs_kmat_inv: np.ndarray) -> np.ndarray:
    """
        kxs = kern(xs, xhist)
        kxs_kmat_inv = kxs @ kmat_inv
    """
    t = yhist.size
    n = kxs_kmat_inv.shape[0]
    yhist = yhist.reshape((t, 1))
    res = kxs_kmat_inv @ yhist
    return res.reshape((n, ))


def mu2_means(yhist: np.ndarray, kxs_kmat_inv: np.ndarray) -> np.ndarray:
    """
        Returns (y1^2, ..., yt^2) kmat_inv @ kxs
        kxs = kern(xs, xhist)
        kxs_kmat_inv = kxs @ kmat_inv
    """
    t = yhist.size
    n = kxs_kmat_inv.shape[0]
    yhist = yhist.reshape((t, 1))
    y2 = np.square(yhist)
    res = kxs_kmat_inv @ y2
    return res.reshape((n, ))


def mvs_est(yhist: np.ndarray, kxs_kmat_inv: np.ndarray, c: float) -> np.ndarray:
    '''
    estimation of MV_c
    '''
    m1 = mu_means(yhist, kxs_kmat_inv)
    m2 = mu2_means(yhist, kxs_kmat_inv)
    return m1 - c * m2 + c * np.square(m1)


def mu_each(x: np.ndarray, xhist: np.ndarray, yhist: np.ndarray,
            kmat_inv: np.ndarray, kern: sk_kern.Kernel) -> np.ndarray:
    t = yhist.size
    kx = kern([x], xhist)
    kx = kx.reshape((t,))
    yhist = yhist.reshape((t, 1))
    ky = kmat_inv @ yhist
    ky = ky.reshape((t,))
    return ky.dot(kx)


def cvars_est(alpha: float, yhist: np.ndarray, kxs_kmat_inv: np.ndarray,
              nu_max: float, nu_min: float) -> np.ndarray:
    t = yhist.size
    n = kxs_kmat_inv.shape[0]
    nus = np.concatenate([yhist, np.array([nu_max, nu_min])])
    # nus = yhist
    nus_mat = np.tile(nus, (t, 1))
    n_nus = nus_mat.shape[1]
    nu_scalr_mat = np.tile(nus, (n, 1))

    ys_mat = np.tile(yhist, (n_nus, 1))
    ys_mat = ys_mat.T
    psi_mat = np.maximum(nus_mat - ys_mat, 0.0)
    kxs_psi = kxs_kmat_inv @ psi_mat
    return np.max(nu_scalr_mat - kxs_psi / alpha, axis=1)


class Parameter:
    def __init__(self, kern: sk_kern.Kernel,
                 B: Optional[float] = None,
                 R: Optional[float] = None,
                 lam=1.0,
                 delta=1e-2,
                 nu_min=None,
                 nu_max=None,
                 num_samples_mv: Optional[int] = None,
                 rhobar: Optional[float] = None,
                 betamv: Optional[float] = None,
                 betamv_var: Optional[float] = None,
                 ):
        self.kern = kern
        self.lam = lam
        self.nu_min = nu_min
        self.nu_max = nu_max
        self.B = B
        self.R = R
        self.delta = delta

        # For RAHBO or MV
        # k in their paper
        self.num_samples_mv = num_samples_mv
        self.rhobar = rhobar
        self.betamv = betamv
        self.betamv_var = betamv_var


class KernBanditBase(ck_base.BanditPolicy):
    def __init__(self, param: Parameter):
        self.param = param
        self.kinv = None

        # 0.5 log det (I + lam^{-1} K_t) = 1/2 sum_{s=1}^{t-x}log(1 + lam^{-1} sigma_{t-1}^2(x_t))
        self.ig = 0.0

        self.xhist = None
        self.yhist = None

    @property
    def kern(self):
        return self.param.kern

    def update(self, selected_arm: np.ndarray, reward: float):
        x_new = np.array([selected_arm])
        y_new = np.array([reward])

        # udpate inverse of kernel matrix and information gain
        if self.xhist is None:
            kmat = self.param.lam + self.kern(x_new, x_new)
            self.kinv = np.linalg.inv(kmat)
            self.ig = math.log(1 + self.kern(x_new, x_new)
                               [0, 0]/self.param.lam) * 0.5
        else:
            s = sigma_each(selected_arm, self.xhist, self.kinv, self.kern)
            self.ig += math.log(1.0 + s**2 / self.param.lam) * 0.5

            b = self.kern(self.xhist, x_new)
            t = self.yhist.size
            b = b.reshape((t,))
            d = self.kern(x_new, x_new)[0, 0] + self.param.lam
            self.kinv = ck_utils.schur_compl_inverse(self.kinv, b, d)

        # update histories
        if self.xhist is None:
            self.xhist = x_new
            self.yhist = y_new
        else:
            self.xhist = np.concatenate([self.xhist, x_new])
            self.yhist = np.concatenate([self.yhist, y_new])


class IGPUCB(KernBanditBase):
    def __init__(self, param: Parameter):
        super().__init__(param)

    def beta(self) -> float:
        a = 2.0 * (self.ig + 1.0 + math.log(1.0 / self.param.delta))
        return self.param.B + self.param.R * math.sqrt(a)

    def select(self, arms: np.ndarray, rng: np.random.Generator) -> np.ndarray:
        if self.xhist is None:
            n = arms.shape[0]
            i = rng.integers(0, n)
        else:
            kxs_kmat_inv = self.kern(arms, self.xhist) @ self.kinv
            ms = mu_means(self.yhist, kxs_kmat_inv)
            ss = sigmas(arms, self.xhist, kxs_kmat_inv, self.kern)
            ucbs = ms + ss * self.beta()
            i = np.argmax(ucbs)
        return arms[i]

    def name(self) -> str:
        return "igpb{:05f}_r{:05f}".format(self.param.B, self.param.R)


class CVKMEUCB(KernBanditBase):
    def __init__(self, param: Parameter, alpha: float):
        super().__init__(param)
        self.alpha = alpha

    def beta(self) -> float:
        a = (2.0 / self.param.lam) * \
            (self.ig + math.log(1.0 / self.param.delta))
        return (self.param.B + 2.0 * math.sqrt(a)) * self.param.R

    def select(self, arms: np.ndarray, rng: np.random.Generator) -> np.ndarray:
        if self.xhist is None:
            n = arms.shape[0]
            i = rng.integers(0, n)
        else:
            kxs_kmat_inv = self.kern(arms, self.xhist) @ self.kinv
            cvars = cvars_est(self.alpha, self.yhist, kxs_kmat_inv,
                              self.param.nu_max, self.param.nu_min)
            ss = sigmas(arms, self.xhist, kxs_kmat_inv, self.kern)
            ss *= self.beta() / self.alpha
            print("cvars_mean", cvars.mean())
            print("ss beta/alpha mean", ss.mean())
            ucbs = cvars + ss
            i = np.argmax(ucbs)
        return arms[i]

    def name(self) -> str:
        return "ckucbb{:05f}_r{:05f}".format(self.param.B, self.param.R)

    def cvar_hat(self, arm: np.ndarray, alpha: float) -> Optional[float]:
        arms = np.array(arm)
        kxs_kmat_inv = self.kern(arms, self.xhist) @ self.kinv
        cvars = cvars_est(alpha, self.yhist, kxs_kmat_inv,
                          self.param.nu_max, self.param.nu_min)
        cvar = float(cvars[0])
        return cvar


class MVKMEUCB(KernBanditBase):
    def __init__(self, param: Parameter, c: float):
        super().__init__(param)
        self.c = c

    # we use constant beta as in RAHBO
    # def beta(self) -> float:
    #     a = (2.0 / self.param.lam) * \
    #         (self.ig + math.log(1.0 / self.param.delta))
    #     return (self.param.B + 2.0 * math.sqrt(a)) * self.param.R

    def select(self, arms: np.ndarray, rng: np.random.Generator) -> np.ndarray:
        if self.xhist is None:
            n = arms.shape[0]
            i = rng.integers(0, n)
        else:
            kxs_kmat_inv = self.kern(arms, self.xhist) @ self.kinv
            mvs = mvs_est(self.yhist, kxs_kmat_inv, self.c)
            ss = sigmas(arms, self.xhist, kxs_kmat_inv, self.kern)
            ucbs = mvs + self.param.betamv * ss + self.param.betamv_var * np.square(ss)
            i = np.argmax(ucbs)
        return arms[i]

    def mv_hat(self, arm: np.ndarray, c: float) -> Optional[float]:
        arms = np.array(arm)
        kxs_kmat_inv = self.kern(arms, self.xhist) @ self.kinv
        mvs = mvs_est(self.yhist, kxs_kmat_inv, c)
        return mvs[0]

    def name(self):
        return "MVUCB_beta{:05f}_betamv{:05f}".format(self.param.betamv, self.param.betamv_var)


def update_kmat_inv(xhist, x_new, kern, kinv, lam) -> np.ndarray:
    b = kern(xhist, x_new)
    t = b.size
    b = b.reshape((t,))
    d = kern(x_new, x_new)[0, 0] + lam
    return ck_utils.schur_compl_inverse(kinv, b, d)


def ig_increment(selected_arm, xhist, kinv, kern, lam) -> float:
    s = sigma_each(selected_arm, xhist, kinv, kern)
    return math.log(1.0 + s**2 / lam) * 0.5


class RAHBO(ck_base.BanditPolicy):
    def __init__(self, param: Parameter, c: float):
        self.c = c
        self.param = param
        self.kinv = None
        self.kinv_var = None

        # self.ig = 0.0
        # self.ig_v = 0.0

        self.xhist = None
        self.yhist = np.array([])

        self.rho_eta = math.sqrt(2) * self.param.rhobar**2 / math.sqrt(self.param.num_samples_mv - 1)
        self.lam_v = self.param.lam + self.rho_eta

        self.observations = np.array([])
        self.vars_hist = np.array([])
        self.t = 0
        self.last_selected_arm = None

    @property
    def kern(self):
        return self.param.kern

    def update(self, selected_arm: np.ndarray, reward: float):
        x_new = np.array([selected_arm])
        k = self.param.num_samples_mv

        self.observations = np.concatenate([self.observations, np.array([reward])])
        self.t += 1
        self.last_selected_arm = selected_arm
        # print("self.t", self.t)

        # udpate inverse of kernel matrix and information gain
        if self.t % k == 0:
            assert self.observations.size == k
            if self.xhist is None:
                csh = self.capital_sigma_hat(x_new)
                kmat = self.param.lam * csh + self.kern(x_new, x_new)
                self.kinv = np.linalg.inv(kmat)

                kmat_v = self.lam_v + self.kern(x_new, x_new)
                self.kinv_var = np.linalg.inv(kmat_v)
                self.xhist = x_new
                self.yhist = np.array([self.observations.mean()])
                self.vars_hist = np.array([variance(self.observations)])
            else:
                self.xhist = np.concatenate([self.xhist, x_new])
                self.yhist = np.concatenate([self.yhist, np.array([self.observations.mean()])])
                self.vars_hist = np.concatenate([self.vars_hist, np.array([variance(self.observations)])])

                self.kinv_var = update_kmat_inv(self.xhist[:-1], x_new, self.kern, self.kinv_var, self.lam_v)

                csh = self.capital_sigma_hat(x_new)
                self.kinv = update_kmat_inv(self.xhist[:-1], x_new, self.kern, self.kinv, self.param.lam * csh)
            self.observations = np.array([])

    def ucb_var(self, x: np.ndarray) -> float:
        if self.yhist.size == 0:
            return self.param.betamv_var
        kxs_kmat_inv = self.kern(x, self.xhist) @ self.kinv_var
        var_mean = mu_means(self.vars_hist, kxs_kmat_inv)[0]
        var_sigma = sigmas(x, self.xhist, kxs_kmat_inv, self.kern)[0]
        return var_mean + self.param.betamv_var * var_sigma

    def capital_sigma_hat(self, x: np.ndarray) -> float:
        return min(self.ucb_var(x), self.param.rhobar**2) / self.param.num_samples_mv

    def var_lcbs(self, xs: np.ndarray) -> np.ndarray:
        kxs_kmat_inv = self.kern(xs, self.xhist) @ self.kinv_var
        var_m = mu_means(self.vars_hist, kxs_kmat_inv)
        var_s = sigmas(xs, self.xhist, kxs_kmat_inv, self.kern)
        return var_m - self.param.betamv_var * var_s

    def m_ucbs(self, xs: np.ndarray) -> np.ndarray:
        kxs_kmat_inv = self.kern(xs, self.xhist) @ self.kinv
        m = mu_means(self.yhist, kxs_kmat_inv)
        s = sigmas(xs, self.xhist, kxs_kmat_inv, self.kern)
        return m + self.param.betamv * s

    def mv_ucbs(self, xs: np.ndarray) -> np.ndarray:
        return self.m_ucbs(xs) - self.c * self.var_lcbs(xs)

    def select(self, arms: np.ndarray, rng: np.random.Generator) -> np.ndarray:
        if self.t == 0:
            n = arms.shape[0]
            i = rng.integers(0, n)
        elif self.t % self.param.num_samples_mv == 0:
            ucbs = self.mv_ucbs(arms)
            i = np.argmax(ucbs)
        else:
            # print("select last arm")
            i = None
        if i is None:
            # print("select last arm", self.last_selected_arm)
            return self.last_selected_arm
        else:
            # print(f"select arm{i}, arm{arms[i]}")
            return arms[i]

    def _mean_est(self, x: np.ndarray) -> float:
        x = np.array(x)
        kxs_kmat_inv = self.kern(x, self.xhist) @ self.kinv
        m = mu_means(self.yhist, kxs_kmat_inv)
        return m[0]

    def _var_est(self, x: np.ndarray) -> float:
        x = np.array(x)
        kxs_kmat_inv = self.kern(x, self.xhist) @ self.kinv_var
        return mu_means(self.vars_hist, kxs_kmat_inv)[0]

    def mv_hat(self, arm: np.ndarray, c: float) -> Optional[float]:
        if self.yhist.size == 0:
            return 0.0
        return self._mean_est(arm) - c * self._var_est(arm)

    def name(self) -> str:
        return "RAHBO_rhobar{:05f}_k{}_beta{:05f}_betav{:05f}".format(
            self.param.rhobar, self.param.num_samples_mv,
            self.param.betamv, self.param.betamv_var)


def variance(obsvs: np.ndarray) -> np.ndarray:
    k = obsvs.size
    m = obsvs.mean()
    d = obsvs - m
    return np.square(d).mean() / (k - 1)
