import ck_bandit.base as ck_base
import numpy as np
import os
import pickle
from typing import List, Optional, Tuple, Dict, Any
from multiprocessing import Pool
import ck_bandit.pols.gp_bandits as gp_bandits
import ck_bandit.utils as ck_utils
import sklearn.gaussian_process.kernels as sk_kern
import ck_bandit.pols.pols_misc as ck_pols_misc
import ck_bandit.envs.synth_envs as ck_synth_envs
import ck_bandit.constants as ck_consts

# There are environments for parameter tuning and evaluation.
# random seed is used in two ways.
# One is to define a random environment and the other is to generate random observation from the environment.
# I use random seeds <= 10 to define environments for tuning parameters and
# random seeds >= 1000 to define environments for evaluation.


def run_res_path_name(pol: ck_base.BanditPolicy, env: ck_base.Environment,
                      nrounds: int, seed: int, alpha: float, mv_c: Optional[float] = None) -> str:
    if mv_c is None:
        fname = "env{}_pol{}_nr{}_sd{}_al{}.pickle".format(
            env.name(), pol.name(), nrounds, seed, alpha)
    else:
        fname = "env{}_pol{}_nr{}_sd{}_al{}_c{}.pickle".format(
            env.name(), pol.name(), nrounds, seed, alpha, mv_c)
    fname = os.path.join("results", fname)
    return fname


def run_policy_and_dump(pol: ck_base.BanditPolicy, env: ck_base.Environment,
                        nrounds: int, rng: np.random.Generator, seed: int, alpha=None, mv_c=None, verbose=True) -> Dict[str, Any]:
    arms = env.arms()
    fname = run_res_path_name(pol, env, nrounds, seed, alpha, mv_c)
    hist = []
    # estimation of CVaR
    cvars_hat = []
    # estimation of MV
    mvs_hat = []
    assert isinstance(pol, ck_pols_misc.RandomPolicy) or pol.xhist is None
    if os.path.exists(fname):
        return
    for t in range(nrounds):
        arm = pol.select(arms, rng)
        r = env.observation(arm, rng)
        pol.update(arm, r)
        hist.append((arm, r))
        if isinstance(pol, gp_bandits.CVKMEUCB):
            cvar_hat = pol.cvar_hat(arm, alpha)
            cvars_hat.append(cvar_hat)
        if isinstance(pol, gp_bandits.MVKMEUCB) or isinstance(pol, gp_bandits.RAHBO):
            mv_hat = pol.mv_hat(arm, mv_c)
            mvs_hat.append(mv_hat)
        if verbose:
            print("round", t)
    if alpha is not None:
        (best_arm, best_reward) = env.best_arm_and_reward(alpha)
        res = {"hist": hist, "best_arm": best_arm,
               "best_reward": best_reward, "alpha": alpha, "cvars_hat": cvars_hat}
    elif mv_c is not None:
        (best_arm, best_reward) = env.best_mv_arm_and_reward(mv_c)
        res = {"hist": hist, "best_arm": best_arm,
               "best_reward": best_reward, "c": mv_c, "mvs_hat": mvs_hat}
    else:
        res = {"hist": hist, "alpha": alpha, "c": mv_c}
    with open(fname, "wb") as fp:
        pickle.dump(res, fp)
    return res


def load_run_res(*args, **kwargs) -> Dict[str, Any]:
    fname = run_res_path_name(*args, **kwargs)
    with open(fname, "rb") as fp:
        return pickle.load(fp)


def run_policies_and_dump(pols: List[ck_base.BanditPolicy], envs: List[ck_base.Environment], nrounds: int,
                          seeds: List[int], nprocs=None):
    repeat = len(pols)
    nrounds_ls = [nrounds for _ in range(repeat)]
    args = list(zip(pols, envs, nrounds_ls, seeds))
    with Pool(nprocs) as p:
        _res = p.starmap(run_policy_and_dump, args)


def param_search_igp(env: ck_base.Environment, kern: sk_kern.Kernel, alpha=None, nrounds=1000):
    seed = 1234
    rng = np.random.default_rng(seed)
    r_params = ck_utils.parameter_candidates(1e-3, 1e-1, 10)
    params = [gp_bandits.Parameter(kern, 0.0, r) for r in r_params]
    pols = (gp_bandits.IGPUCB(p) for p in params)
    for pol in pols:
        args = [(pol, env, nrounds, rng, seed, alpha)]
        # to clear memory run function in a different process.
        with Pool(1) as p:
            p.starmap(run_policy_and_dump, args)


def param_search_cvarucb(env: ck_base.Environment, kern: sk_kern.Kernel, alpha: float, nrounds=1000):
    seed = 1234
    rng = np.random.default_rng(seed)
    r_params = ck_utils.parameter_candidates(1e-3, 1e-1, 10)
    nu_min, nu_max = env.min_max_of_q(1e-5)
    params = [gp_bandits.Parameter(
        kern, 0.0, r, nu_min=nu_min, nu_max=nu_max) for r in r_params]
    pols = (gp_bandits.CVKMEUCB(p, alpha) for p in params)
    # to clear memory run function in a different process.
    for pol in pols:
        args = [(pol, env, nrounds, rng, seed, alpha)]
        with Pool(1) as p:
            p.starmap(run_policy_and_dump, args)


def run_rand_in_paramsearch_env(env: ck_base.Environment, alpha=None, nrounds=1000):
    seed = 1234
    pol = ck_pols_misc.RandomPolicy()
    rng = np.random.default_rng(seed)
    run_policy_and_dump(pol, env, nrounds, rng, seed, alpha)


DIM_KERNEL_MAP = {
    1: sk_kern.Matern(0.2),
    2: sk_kern.Matern(0.2),
    # 0.2 is too dificulut for small time interval
    3: sk_kern.Matern(ck_consts.LS, nu=ck_consts.MTN_NU)
}

KERN_NAME = {
    1: "mtn0.2",
    2: "mtn0.2",
    3: "mtn%s_%s" % (ck_consts.MTN_NU, ck_consts.LS),
}


DIM_NPTS_EACH_MAP = {
    1: 30,
    2: 30,
    3: 10,
}


def normal_env_for_param_search(dim=1, seed=0) -> ck_base.Environment:
    assert seed <= 10
    kern = DIM_KERNEL_MAP[dim]
    npts_each_dim = DIM_NPTS_EACH_MAP[dim]
    rng = np.random.default_rng(seed)
    env = ck_synth_envs.RandNormalEnv(
        kern, rng, seed, dim, KERN_NAME[dim], npts_each_dim=npts_each_dim)
    return env


def normal_env_for_eval(dim=1, seed=1000) -> ck_base.Environment:
    assert seed >= 1000
    kern = DIM_KERNEL_MAP[dim]
    npts_each_dim = DIM_NPTS_EACH_MAP[dim]
    rng = np.random.default_rng(seed)
    env = ck_synth_envs.RandNormalEnv(
        kern, rng, seed, dim, KERN_NAME[dim], npts_each_dim=npts_each_dim)
    return env


def lognormal_env_for_param_search(dim=1, seed=0) -> ck_base.Environment:
    assert seed <= 10
    kern = DIM_KERNEL_MAP[dim]
    npts_each_dim = DIM_NPTS_EACH_MAP[dim]
    rng = np.random.default_rng(seed)
    env = ck_synth_envs.RandLogNormalEnv(
        kern, rng, seed, dim, KERN_NAME[dim], npts_each_dim=npts_each_dim)
    return env


def lognormal_env_for_eval(dim=1, seed=1000) -> ck_base.Environment:
    assert seed >= 1000
    kern = DIM_KERNEL_MAP[dim]
    npts_each_dim = DIM_NPTS_EACH_MAP[dim]
    rng = np.random.default_rng(seed)
    env = ck_synth_envs.RandLogNormalEnv(
        kern, rng, seed, dim, KERN_NAME[dim], npts_each_dim=npts_each_dim)
    return env


def cumreg_cvars_using_env(pol, env, nrounds, seed, alpha, norm_fac=1.0) -> np.ndarray:
    try:
        d = load_run_res(pol, env, nrounds, seed, alpha)
    except FileNotFoundError:
        d = load_run_res(pol, env, nrounds, seed, None)
    selected_arms = [x for x, _ in d["hist"]]
    selected_arms = np.array(selected_arms)
    # args = [(arms, alpha)]
    selected_cvars = np.array(env.cvars(selected_arms, alpha))
    # with Pool(1) as p:
    #     cvars = p.starmap(env.cvars, args)
    #     cvars = cvars[0]
    best_cvar = env.best_arm_and_reward(alpha)[1]
    # with Pool(1) as p:
    #     res = p.starmap(env.best_arm_and_reward, [(alpha,)])
    #     _, best_cvar = res[0]
    regrets = best_cvar - selected_cvars
    creg = np.array([regrets[:t].sum() for t in range(1, nrounds + 1)])
    return creg / norm_fac


def cumreg_mvs_using_env(pol, env, nrounds, seed, mv_c, norm_fac=1.0) -> np.ndarray:
    alpha = None
    try:
        d = load_run_res(pol, env, nrounds, seed, alpha, mv_c)
    except FileNotFoundError:
        d = load_run_res(pol, env, nrounds, seed, alpha, None)
    selected_arms = np.array([x for x, _ in d['hist']])
    selected_mvs = np.array(env.mean_vars(selected_arms, mv_c))
    best_mv = env.best_mv_arm_and_reward(mv_c)[1]
    regrets = best_mv - selected_mvs
    creg = np.array([regrets[:t].sum() for t in range(1, nrounds + 1)])
    return creg / norm_fac


def best_param_using_true_cvar(pols: List[gp_bandits.KernBanditBase], envs: List[ck_base.Environment],
                               nrounds, alpha, seed=1234) -> gp_bandits.Parameter:
    # rand_pol = ck_pols_misc.RandomPolicy()
    scores = []
    for pol in pols:
        scores_pol = []
        for env in envs:
            # rand_res = run_pol.cumreg_cvars_using_env(rand_pol, env, nrounds, seed, alpha)
            # norm_fac = rand_res[nrounds - 1]
            cum_reg = cumreg_cvars_using_env(
                pol, env, nrounds, seed, alpha, norm_fac=1)
            scores_pol.append(cum_reg[nrounds - 1])
        scores.append(np.array(scores_pol).mean())
    idx = np.argmin(scores)
    return pols[idx].param


def best_param_using_cvarhat(pols: List[gp_bandits.CVKMEUCB], envs: List[ck_base.Environment], nrounds, alpha, seed=1234) -> gp_bandits.Parameter:
    scores = []
    for pol in pols:
        scores_pol = []
        for env in envs:
            d = load_run_res(pol, env, nrounds, seed, alpha)
            scores_pol.append(np.mean(d["cvars_hat"]))
        scores_pol = np.array(scores_pol)
        scores.append(scores_pol.mean())
    idx = np.argmax(scores)
    return pols[idx].param


def best_param_using_mvhat(pols: List[ck_base.BanditPolicy], envs: List[ck_base.Environment],
                           nrounds, mv_c, seed=ck_consts.PARAM_SEARCH_OBS_SEED):
    scores = []
    for pol in pols:
        scores_pol = []
        for env in envs:
            d = load_run_res(pol, env, nrounds, seed, None, mv_c=mv_c)
            scores_pol.append(np.mean(d["mvs_hat"]))
        scores_pol = np.array(scores_pol)
        scores.append(scores_pol.mean())
    idx = np.argmax(scores)
    return pols[idx].param


def best_param_using_observation(pols: List[gp_bandits.KernBanditBase], envs: List[ck_base.Environment],
                                 nrounds, alpha=None, seed=1234) -> gp_bandits.Parameter:
    scores = []
    for pol in pols:
        scores_pol = []
        for env in envs:
            d = load_run_res(pol, env, nrounds, seed, alpha)
            obs = np.array([y for x, y in d["hist"]])
            scores_pol.append(np.mean(obs))
        scores_pol = np.array(scores_pol)
        scores.append(scores_pol.mean())
    idx = np.argmax(scores)
    return pols[idx].param


ENV_FNS_PARAM_SEARCH = {
    "normal": normal_env_for_param_search,
    "log_normal": lognormal_env_for_param_search,
}

ENV_FNS_EVAL = {
    "normal": normal_env_for_eval,
    "log_normal": lognormal_env_for_eval,
}
