import numpy as np
import sklearn.gaussian_process.kernels as sk_kern
from typing import Tuple
import math


def schur_compl_inverse(ainv: np.ndarray, b: np.ndarray, d: float) -> np.ndarray:
    '''
    let m = [[a, b], [b^t, d]] be a symmetric matrix.
    returns a inverse of m using ainv and Schur complement.
    '''
    (n, _) = ainv.shape
    b = b.reshape((b.size,))
    res = np.zeros((n + 1, n + 1))
    ainvb = ainv.dot(b)
    p = d - ainvb.dot(b)
    pinv = p**(-1)
    res[:n, :n] = ainv + pinv * outerprod(ainvb)
    res[n, n] = pinv
    res[:n, n] = - ainvb * pinv
    res[n, :n] = - ainvb * pinv
    return res


def outerprod(x: np.ndarray) -> np.ndarray:
    n = x.size
    a = x.reshape((n, 1))
    return a @ a.T


def dots_of_arrays(xs: np.ndarray, ys: np.ndarray) -> np.ndarray:
    return np.multiply(xs, ys).sum(1)


def random_func_rkhs_of_unit_norm(num_pts: int, kern: sk_kern.Kernel, lbs: np.ndarray,
                                  ubs: np.ndarray, rng=None, coeffs_lb=-1.0, coeffs_ub=1.0) -> Tuple[np.ndarray, np.ndarray]:
    '''
    It is assumed that points are generated from a region defined by lbs and ubs.
    Returns a tuple fo (coeffs, pts)
    '''
    if rng is None:
        rng = np.random.default_rng()
    dim = lbs.shape[0]
    pts = rng.uniform(lbs, ubs, size=(num_pts, dim))
    coeffs = rng.uniform(coeffs_lb, coeffs_ub, size=(num_pts, 1))
    K = kern(pts, pts)
    norm = (K @ coeffs).reshape((num_pts,)).dot(coeffs.reshape((num_pts,)))
    norm = math.sqrt(norm)
    return (coeffs/norm, pts)


def parameter_candidates(lb: float, ub: float, num: int) -> np.ndarray:
    return np.exp(np.linspace(np.log(lb), np.log(ub), num))
