import unittest

import ck_bandit.run_policy as ck_run_pol
import ck_bandit.pols.gp_bandits as gp_bandits
import ck_bandit.envs.synth_envs as ck_synth_envs
import sklearn.gaussian_process.kernels as sk_kern
import ck_bandit.utils as ck_utils
import numpy as np
import ck_bandit.pols.pols_misc as ck_pols_misc
from multiprocessing import Pool
import pickle
from ck_bandit.constants import *
from typing import List, Any


def params_for_search_igp(kern) -> List[Any]:
    r_params = ck_utils.parameter_candidates(1e-2, 1.0, 10)
    params = [gp_bandits.Parameter(kern, 0.0, r) for r in r_params]
    return params


def params_for_search_cvar(kern) -> List[Any]:
    r_params = ck_utils.parameter_candidates(1e-3, 1e-1, 10)
    params = [gp_bandits.Parameter(kern, 0.0, r) for r in r_params]
    return params


def run_pol_and_dump_wrapper_prams(
        dim: int,
        env_type: str,
        env_seed: int,
        pol: gp_bandits.KernBanditBase,
        alpha: float,
        fix_param: bool = True):
    obs_seed = PARAM_SEARCH_OBS_SEED
    rng = np.random.default_rng(obs_seed)
    env = ck_run_pol.ENV_FNS_PARAM_SEARCH[env_type](dim=dim, seed=env_seed)
    nrounds = 200
    if fix_param:
        set_nu_minmax(pol, env)
    ck_run_pol.run_policy_and_dump(pol, env, nrounds, rng=rng, seed=obs_seed, alpha=alpha)


class TestParamSearch(unittest.TestCase):

    def test_param_search_pols(self):
        envs_types = ENV_TYPES
        for dim in DIMS:
            for env_type in envs_types:
                self.test_rand(dim, env_type)
                self.test_cvarucb(dim, env_type)
                self.test_igp(dim, env_type)

    def test_rand(self, dim, env_type):
        for env_seed in PARAM_SEARCH_SEEDS:
            pol = ck_pols_misc.RandomPolicy()
            alpha = None
            args = (dim, env_type, env_seed, pol, alpha, False)
            with Pool(1) as p:
                p.apply(run_pol_and_dump_wrapper_prams, args)

    def test_igp(self, dim, env_type):
        kern = ck_run_pol.DIM_KERNEL_MAP[dim]
        params = params_for_search_igp(kern)
        alpha = None
        for env_seed in PARAM_SEARCH_SEEDS:
            for prm in params:
                pol = gp_bandits.IGPUCB(prm)
                args = (dim, env_type, env_seed, pol, alpha, False)
                with Pool(1) as p:
                    p.apply(run_pol_and_dump_wrapper_prams, args)

    def test_cvarucb(self, dim, env_type):
        kern = ck_run_pol.DIM_KERNEL_MAP[dim]
        params = params_for_search_cvar(kern)
        for env_seed in PARAM_SEARCH_SEEDS:
            for alpha in ALPHAS:
                for prm in params:
                    pol = gp_bandits.CVKMEUCB(prm, alpha)
                    args = (dim, env_type, env_seed, pol, alpha, True)
                    with Pool(1) as p:
                        p.apply(run_pol_and_dump_wrapper_prams, args)

    def test_best_igp_cvar(self):
        alpha = 0.1
        pi = best_igp_using_true_cvar(alpha, 3)
        print("igp", pi.R)
        pc = best_cvarucb_using_cvarhat(alpha, 3)
        print("cvar", pc.R)


ENV_TYPES = ["normal", "log_normal"]


def best_igp_using_true_cvar(alpha, dim, env_type, nrounds=200) -> gp_bandits.Parameter:
    assert env_type in ENV_TYPES
    kern = ck_run_pol.DIM_KERNEL_MAP[dim]
    params = params_for_search_igp(kern)
    pols = [gp_bandits.IGPUCB(p) for p in params]
    seeds = PARAM_SEARCH_SEEDS
    ENV_FNS = {
        "normal": ck_run_pol.normal_env_for_param_search,
        "log_normal": ck_run_pol.lognormal_env_for_param_search
    }
    envs = [ENV_FNS[env_type](dim=dim, seed=seed) for seed in seeds]
    p = ck_run_pol.best_param_using_true_cvar(pols, envs, nrounds=nrounds, alpha=alpha)
    return p


def best_igp_using_obs_sum(_alpha, dim, env_type, nrounds=200) -> gp_bandits.Parameter:
    kern = ck_run_pol.DIM_KERNEL_MAP[dim]
    params = params_for_search_igp(kern)
    pols = [gp_bandits.IGPUCB(p) for p in params]
    envs = [ck_run_pol.ENV_FNS_PARAM_SEARCH[env_type](dim, seed=seed) for seed in PARAM_SEARCH_SEEDS]
    p = ck_run_pol.best_param_using_observation(pols, envs, nrounds=nrounds)
    return p


def best_cvarucb_using_cvarhat(alpha, dim, env_type, nrounds=200) -> gp_bandits.Parameter:
    assert env_type in ENV_TYPES
    kern = ck_run_pol.DIM_KERNEL_MAP[dim]
    params = params_for_search_cvar(kern)
    pols = [gp_bandits.CVKMEUCB(p, alpha) for p in params]
    seeds = PARAM_SEARCH_SEEDS

    ENV_FNS = {
        "normal": ck_run_pol.normal_env_for_param_search,
        "log_normal": ck_run_pol.lognormal_env_for_param_search
    }

    envs = [ENV_FNS[env_type](dim=dim, seed=seed) for seed in seeds]
    p = ck_run_pol.best_param_using_cvarhat(pols, envs, nrounds=nrounds, alpha=alpha)
    return p


def load_best_params():
    with open(BEST_PARAMS_PATH, "rb") as fp:
        return pickle.load(fp)


def set_nu_minmax(pol, env):
    nu_min, nu_max = env.min_max_of_q(1e-5)
    pol.param.nu_max, pol.param.nu_min = nu_max, nu_min


def run_pol_and_dump_wrapper_eval(
        dim,
        env_type,
        env_seed,
        pol: gp_bandits.KernBanditBase,
        nrounds,
        obs_seed,
        alpha,
        fix_param=True):
    # env_seedは環境を決めるため、obs_seedはobservationを決めるため。
    rng = np.random.default_rng(obs_seed)
    env = ck_run_pol.ENV_FNS_EVAL[env_type](dim=dim, seed=env_seed)
    if fix_param:
        # We assume that hig-prob support is known
        set_nu_minmax(pol, env)
    ck_run_pol.run_policy_and_dump(pol, env, nrounds, rng=rng, seed=obs_seed, alpha=alpha)


class TestEvalSynthEnvs(unittest.TestCase):
    def test_rand(self, dim, env_type):
        alpha = None
        for env_seed in EVAL_SEEDS:
            pol = ck_pols_misc.RandomPolicy()
            args = (dim, env_type, env_seed, pol, NROUNDS_EVAL, EVAL_OBS_SEED, alpha, False)
            with Pool(1) as p:
                p.apply(run_pol_and_dump_wrapper_eval, args)

    def dump_best_params(self):
        dims = DIMS
        env_typs = ENV_TYPES
        d = {}
        d_igp = {(dim, env_type, alpha): best_igp_using_obs_sum(alpha, dim, env_type=env_type)
                 for dim in dims for env_type in env_typs for alpha in ALPHAS}

        d_igp_true = {(dim, env_type, alpha): best_igp_using_true_cvar(alpha, dim, env_type=env_type)
                      for dim in dims for env_type in env_typs for alpha in ALPHAS}

        d_cvarucb = {(dim, env_type, alpha): best_cvarucb_using_cvarhat(alpha, dim, env_type=env_type)
                     for dim in dims for env_type in env_typs for alpha in ALPHAS}
        d["igp"] = d_igp
        d["cvarucb"] = d_cvarucb
        d["igp-true-cvar"] = d_igp_true
        with open(BEST_PARAMS_PATH, "wb") as fp:
            pickle.dump(d, fp)

    def test_igp(self, dim, env_type, eval_true_cvar=False):
        d = load_best_params()
        for alpha in ALPHAS:
            prm = d["igp"][(dim, env_type, alpha)]
            prm_true = d["igp-true-cvar"][(dim, env_type, alpha)]
            for env_seed in EVAL_SEEDS:
                with Pool(1) as p:
                    pol = gp_bandits.IGPUCB(prm)
                    args = (dim, env_type, env_seed, pol, NROUNDS_EVAL, EVAL_OBS_SEED, alpha)
                    p.apply(run_pol_and_dump_wrapper_eval, args)
                if eval_true_cvar:
                    with Pool(1) as p:
                        pol = gp_bandits.IGPUCB(prm_true)
                        args = (dim, env_type, env_seed, pol, NROUNDS_EVAL, EVAL_OBS_SEED, alpha)
                        p.apply(run_pol_and_dump_wrapper_eval, args)

    def test_cvarucb(self, dim, env_type):
        d = load_best_params()
        for alpha in ALPHAS:
            prm = d["cvarucb"][(dim, env_type, alpha)]
            for env_seed in EVAL_SEEDS:
                with Pool(1) as p:
                    pol = gp_bandits.CVKMEUCB(prm, alpha)
                    args = (dim, env_type, env_seed, pol, NROUNDS_EVAL, EVAL_OBS_SEED, alpha)
                    p.apply(run_pol_and_dump_wrapper_eval, args)

    def test_eval_all_pols(self):
        self.dump_best_params()
        env_typs = ENV_TYPES
        for dim in DIMS:
            for env_type in env_typs:
                self.test_rand(dim, env_type)
                self.test_igp(dim, env_type)
                self.test_cvarucb(dim, env_type)


if __name__ == '__main__':
    unittest.main()
