import unittest

import ck_bandit.run_policy as ck_run_pol
import ck_bandit.base as ck_base
import ck_bandit.pols.gp_bandits as gp_bandits
import ck_bandit.envs.synth_envs as ck_synth_envs
import sklearn.gaussian_process.kernels as sk_kern
import ck_bandit.utils as ck_utils
import numpy as np
import ck_bandit.pols.pols_misc as ck_pols_misc
from multiprocessing import Pool
import pickle
from ck_bandit.constants import *
from typing import List, Any, Optional


def params_for_mvucb(kern) -> List[Any]:
    '''
    A candidate list of parameters for MVPKEUCB
    '''
    betas = ck_utils.parameter_candidates(1e-2, 1.0, 4)
    betas_v = ck_utils.parameter_candidates(1e-2, 1.0, 4)
    return [gp_bandits.Parameter(kern, betamv=beta, betamv_var=betav)
            for beta in betas for betav in betas_v]


def params_for_rahbo(kern, num_samples_mv=5) -> List[Any]:
    '''
    A candidate list of parameters for RAHBO 
    '''
    betas = ck_utils.parameter_candidates(1e-2, 1.0, 4)
    betas_v = ck_utils.parameter_candidates(1e-2, 1.0, 4)
    # To ease parameter search, we set rhobar = 1.0
    return [gp_bandits.Parameter(kern, rhobar=1.0,
                                 num_samples_mv=num_samples_mv, betamv=beta, betamv_var=betav)
            for beta in betas for betav in betas_v]


def run_pol_and_dump_wrapper_param_search(
    dim: int,
    env_type: str,
    env_seed: int,
    pol: ck_base.BanditPolicy,
    c: float,
):
    '''
    Used for parameter search.
    '''
    obs_seed = PARAM_SEARCH_OBS_SEED
    rng = np.random.default_rng(obs_seed)
    env = ck_run_pol.ENV_FNS_PARAM_SEARCH[env_type](dim=dim, seed=env_seed)
    nrounds = 200
    ck_run_pol.run_policy_and_dump(pol, env, nrounds, rng=rng, seed=obs_seed, mv_c=c)


def run_pol_and_dump_wrapper_eval(
        dim: int,
        env_type: str,
        env_seed: int,
        pol: ck_base.BanditPolicy,
        nrounds: int,
        obs_seed: int,
        c: Optional[int]):
    rng = np.random.default_rng(obs_seed)
    env = ck_run_pol.ENV_FNS_EVAL[env_type](dim=dim, seed=env_seed)
    ck_run_pol.run_policy_and_dump(pol, env, nrounds, rng=rng, seed=obs_seed, mv_c=c)


class TestParamSearch(unittest.TestCase):

    def test_rand(self, dim, env_type):
        for env_seed in PARAM_SEARCH_SEEDS:
            pol = ck_pols_misc.RandomPolicy()
            c = None
            args = (dim, env_type, env_seed, pol, c)
            with Pool(1) as p:
                p.apply(run_pol_and_dump_wrapper_param_search, args)

    def test_mvucb(self, dim, env_type):
        kern = ck_run_pol.DIM_KERNEL_MAP[dim]
        params = params_for_mvucb(kern)
        for env_seed in PARAM_SEARCH_SEEDS:
            for c in MV_CS:
                for prm in params:
                    pol = gp_bandits.MVKMEUCB(prm, c)
                    args = (dim, env_type, env_seed, pol, c)
                    with Pool(1) as p:
                        p.apply(run_pol_and_dump_wrapper_param_search, args)

    def test_rahbo(self, dim, env_type):
        kern = ck_run_pol.DIM_KERNEL_MAP[dim]
        params = params_for_rahbo(kern)
        for env_seed in PARAM_SEARCH_SEEDS:
            for c in MV_CS:
                for prm in params:
                    pol = gp_bandits.RAHBO(prm, c)
                    args = (dim, env_type, env_seed, pol, c)
                    with Pool(1) as p:
                        p.apply(run_pol_and_dump_wrapper_param_search, args)

    def test_param_search_pols(self):
        env_types = ENV_TYPES
        for dim in DIMS:
            for env_type in env_types:
                self.test_rand(dim, env_type)
                self.test_mvucb(dim, env_type)
                self.test_rahbo(dim, env_type)


def best_mvucb_using_mvhat(c, dim, env_type, nrounds=200) -> gp_bandits.Parameter:
    assert env_type in ENV_TYPES
    kern = ck_run_pol.DIM_KERNEL_MAP[dim]
    params = params_for_mvucb(kern)
    pols = [gp_bandits.MVKMEUCB(p, c) for p in params]
    seeds = PARAM_SEARCH_SEEDS
    envs = [ck_run_pol.ENV_FNS_PARAM_SEARCH[env_type](dim=dim, seed=seed) for seed in seeds]
    p = ck_run_pol.best_param_using_mvhat(pols, envs, nrounds, c)
    return p


def best_rahbo_using_mvhat(c, dim, env_type, nrounds=200) -> gp_bandits.Parameter:
    kern = ck_run_pol.DIM_KERNEL_MAP[dim]
    params = params_for_rahbo(kern)
    pols = [gp_bandits.RAHBO(p, c) for p in params]
    seeds = PARAM_SEARCH_SEEDS
    envs = [ck_run_pol.ENV_FNS_PARAM_SEARCH[env_type](dim=dim, seed=seed) for seed in seeds]
    p = ck_run_pol.best_param_using_mvhat(pols, envs, nrounds, c)
    return p


def load_best_params():
    with open(BEST_PARAMS_MV_PATH, 'rb') as fp:
        return pickle.load(fp)


class TestEvalSynthEnvs(unittest.TestCase):
    def test_rand(self, dim, env_type):
        c = None
        for env_seed in EVAL_SEEDS:
            pol = ck_pols_misc.RandomPolicy()
            args = (dim, env_type, env_seed, pol, NROUNDS_EVAL, EVAL_OBS_SEED, c)
            with Pool(1) as p:
                p.apply(run_pol_and_dump_wrapper_eval, args)

    def dump_best_params(self):
        dims = DIMS
        env_types = ENV_TYPES
        res = {}
        d_mvucb = {(dim, env_type, c): best_mvucb_using_mvhat(c, dim, env_type)
                   for dim in dims for env_type in env_types for c in MV_CS}
        d_rahbo = {(dim, env_type, c): best_rahbo_using_mvhat(c, dim, env_type)
                   for dim in dims for env_type in env_types for c in MV_CS}
        res['mv_ucb'] = d_mvucb
        res['rahbo'] = d_rahbo
        with open(BEST_PARAMS_MV_PATH, 'wb') as fp:
            pickle.dump(res, fp)

    def test_mvucb(self, dim, env_type):
        d = load_best_params()
        for c in MV_CS:
            prm = d["mv_ucb"][(dim, env_type, c)]
            for env_seed in EVAL_SEEDS:
                with Pool(1) as p:
                    pol = gp_bandits.MVKMEUCB(prm, c)
                    args = (dim, env_type, env_seed, pol, NROUNDS_EVAL, EVAL_OBS_SEED, c)
                    p.apply(run_pol_and_dump_wrapper_eval, args)

    def test_rahbo(self, dim, env_type):
        d = load_best_params()
        for k in NUM_SAMPLES_LS:
            for c in MV_CS:
                prm = d["rahbo"][(dim, env_type, c)]
                prm.num_samples_mv = k
                for env_seed in EVAL_SEEDS:
                    with Pool(1) as p:
                        pol = gp_bandits.RAHBO(prm, c)
                        args = (dim, env_type, env_seed, pol, NROUNDS_EVAL, EVAL_OBS_SEED, c)
                        p.apply(run_pol_and_dump_wrapper_eval, args)

    def test_eval_all_pols(self):
        self.dump_best_params()
        env_types = ENV_TYPES
        for dim in DIMS:
            for env_type in env_types:
                self.test_rand(dim, env_type)
                self.test_rahbo(dim, env_type)
                self.test_mvucb(dim, env_type)
